#include <stdio.h>
#include <ctype.h>

#include "sim.h"
#include "memory.h"

//------------------------------------------------------
// writeByte(mem, address, value)
//    Stores the word (value) at the byte address given.
//    Returns 0 on success.
//------------------------------------------------------

int writeByte(byte mem[], unsigned int address, byte value) {
  mem[address & MEMADDRMASK] = value;
  return 0;
}

//------------------------------------------------------
// fetchByte(mem, address)
//    Returns the the byte at the address given.
//------------------------------------------------------

byte fetchByte(byte mem[], unsigned int address) {
  return mem[address & MEMADDRMASK];
}

//------------------------------------------------------
// writeWord(mem, address, value)
//    Stores the word (value) at the byte address given.
//    Implements little endian.
//    Returns 0 on success.
//------------------------------------------------------

int writeWord(byte mem[], unsigned int address, int value) {
  int b;
  for ( b=0; b<4; b++ ) {
    mem[address++ & MEMADDRMASK] = value & 0xff;
    value >>= 8;
  }

  return 0;
}

//------------------------------------------------------
// fetchWord(mem, address)
//    Returns the word (value) at the byte address given.
//    Implements little endian.
//------------------------------------------------------

int fetchWord(byte mem[], unsigned int address) {
  int b;
  int result = 0;
  for ( b=0; b<4; b++ ) {
    result |= mem[address++ & MEMADDRMASK] << (b*8);
  }
  return result;
}


//----------------------------------------------------------------------
// load(byte mem[])
//   Loads hex input into mem, reading from stdin until EOF.
//   Comments in input denoted by '#', and run to end of line.
//   Returns 0 for success, non-zero for error.
//----------------------------------------------------------------------

#define IDLE 0
#define HALF 1
#define COMMENT 2

int runLoader(byte mem[]) {
  int line = 1;
  unsigned int nextAddress;
  int state = IDLE;
  byte nextByte = 0;
  char c;

  //----------------------------
  // uninitialize memory
  //----------------------------
  for (nextAddress=0; nextAddress<MEMSIZE; nextAddress++ ) writeByte( mem, nextAddress, 0xCD );

  //----------------------------------
  // now load user input (from stdin)
  //----------------------------------
  nextAddress = 0;
  while ( (c=getchar()) != EOF ) {
    
    if ( state == IDLE || state == HALF) {

      if ( c == '@' ) {
	if ( state != IDLE ) {
	  printf("Line %d: Odd number of hex digits precede @\n", line);
	  return 1;
	}
	if ( 1 != scanf("%x", &nextAddress) ) {
	  printf("Line %d: Didn't understand @ directive\n", line) ;
	  return 1;
	}
      }

      else if ( c >= '0' && c <= '9' )  {
	if ( state == IDLE ) {
	  nextByte = (c-'0') << 4;
	  state = HALF;
	} else {
	  nextByte |= c - '0';
	  writeByte( mem, nextAddress++, nextByte );
	  nextByte = 0;
	  state = IDLE;
	}
      }

      else if ( toupper(c) >= 'A' && toupper(c) <= 'F') { 
	if ( state == IDLE ) {
	  nextByte = (toupper(c)-'A' + 10) << 4;
	  state = HALF;
	} else {
	  nextByte |= toupper(c)-'A' + 10;
	  writeByte( mem, nextAddress++, nextByte );
	  nextByte = 0;
	  state = IDLE;
	}
      }

      else if ( c == '#' ) {
	if ( state == HALF ) {
	  printf("Line %d: Odd number of hex digits precede comment\n", line);
	  return 1;
	}
	state = COMMENT;
      }

      else if ( !isspace(c) ) {
	printf("Line %d: Illegal character in input: '%c'\n", line, c );
	return 1;
      }
    }

    else if ( state == COMMENT && c == '\n' ) state = IDLE;

    if ( c == '\n' ) {
      if ( state == HALF ) {
	printf("Line %d: Odd number of hex digits given\n", line);
	return 1;
      }
      line++;
    }

  }

  if ( state == HALF ) {
    printf("Line %d: Odd number of hex digits given\n", line);
    return 1;
  }

  return 0;
}
